"""http_headers"""


import dash


def setup_application_http_response_headers(dash_app: dash.Dash):
    """Add HTTP headers to the response"""
    server = dash_app.server

    @server.after_request
    def add_headers(response):
        response.headers.add(
            "Content-Security-Policy",
            "default-src 'self' 'unsafe-eval' 'unsafe-inline' data:",
        )
        response.headers.add("X-Content-Type-Options", "nosniff")
        response.headers.add("X-Frame-Options", "DENY")
        response.headers.add("Referrer-Policy", "no-referrer")
        # The below header will disable all browser features for the dashboard website
        response.headers.add(
            "Permission-Policy",
            "accelerometer=(), ambient-light-sensor=(), autoplay=(), battery=(), camera=(), "
            "cross-origin-isolated=(), display-capture=(), document-domain=(), encrypted-media=()"
            ", execution-while-not-rendered=(), execution-while-out-of-viewport=(), fullscreen=(),"
            " geolocation=(), gyroscope=(), keyboard-map=(), magnetometer=(), microphone=(), "
            "midi=(), navigation-override=(), payment=(), picture-in-picture=(), "
            "publickey-credentials-get=(), screen-wake-lock=(), sync-xhr=(), usb=(), web-share=()"
            ", xr-spatial-tracking=()",
        )
        return response



"""filter_panel"""
from typing import Optional
from dash import html

from gov_uk_dashboards.components.plotly.heading import heading2
from .collapsible_panel import collapsible_panel
from .row_component import row_component


def filter_panel(
    children, title: Optional[str] = None, sub_title: str = "Select and filter metrics"
):
    """
    A container with a grey background and a title that allows the user to select and filter metrics
    on the dashboard.

    Args:
        children (list): Dash HTML elements representing the individual selection
            and filtering widgets. E.g.dropdown menus, sliders, text input boxes etc.
        title (Optional[str]): Title for filter panel
        sub_title (str): sub_title for filter panel (default is 'Select and filter metrics')
    """
    return row_component(
        html.Div(
            [
                heading2(title) if title else None,
                collapsible_panel(
                    title=sub_title,
                    default_open=True,
                    children=[
                        html.Div(
                            html.A(
                                "Clear all selections",
                                className="govuk-button govuk-button--warning govuk-!-margin-0",
                                href="?",
                                role="button",
                            ),
                            style={"width": "100%"},
                        ),
                        *children,
                    ],
                ),
            ],
            className="govuk-!-margin-0",
            style={"flex": "1 1 100%"},
        )
    )


def hidden_filter(html_id):
    """
    An empty, invisible HTML element that stands in for a filter component.

    Used to keep the state of the filter panel current across dashboards,
    even those that don't allow the user to filter for a particular metric.
    """
    return html.Div(id=html_id, style={"display": "none"})



"""warning_text"""
from dash import html


def warning_text(text: str):
    """
    Return Gov UK Design component warning text component, ! with text.
    """
    text = html.Div(
        [
            html.Span(
                ["!"], className="govuk-warning-text__icon", **{"aria-hidden": "true"}
            ),
            html.Strong(
                [
                    html.Span(["Warning"], className="govuk-warning-text__assistive"),
                    text,
                ],
                className="govuk-warning-text__text",
            ),
        ],
        className="govuk-warning-text",
    )
    return text



"""
Scripts for deploying dashboard code to the DLUHC DAP Hosting Area.
The DAP Hosting area is maintained by the DLUHC DAP Support team.
"""
import subprocess
import tempfile


def deploy_main_branch_to_pydash_test_environment(
    github_repo: str, dap_repo: str, dap_repo_folder_name: str
):
    """
    Deploying to the DLUHC DAP hosting area involves pushing the dashboards code
    from the GitHub main branch to the pydash repos test branch

    See also our internal documentation section "DAP Hosting environment (pydash)"
    https://github.com/communitiesuk/plotly_dashboard_docs#dap-hosting-environment-pydash
    """
    with tempfile.TemporaryDirectory() as temporary_cloning_directory:

        pydash_repo = _clone_repo(
            url=dap_repo,
            into=temporary_cloning_directory,
            path_to_code=dap_repo_folder_name,
        )

        _add_git_remote(
            url=github_repo,
            name="github",
            repo_path=pydash_repo,
        )

        _update_git_reference(
            branch="Test",
            to_point_at="github/main",
            repo_path=pydash_repo,
        )

        _force_push_changes_to_test_branch_of_pydash_repo(
            branch="Test", repo_path=pydash_repo
        )


def _clone_repo(url: str, path_to_code: str, into: str) -> str:
    subprocess.check_call(
        ["git", "clone", url],
        cwd=into,
    )
    return f"{into}\\{path_to_code}"


def _add_git_remote(url: str, name: str, repo_path: str):
    subprocess.check_call(
        [
            "git",
            "remote",
            "add",
            name,
            url,
        ],
        cwd=repo_path,
    )
    subprocess.check_call(["git", "fetch", name], cwd=repo_path)


def _update_git_reference(branch: str, to_point_at: str, repo_path: str):
    _checkout_branch(repo_path, branch=branch)

    subprocess.check_call(
        ["git", "reset", "--hard", to_point_at],
        cwd=repo_path,
    )


def _checkout_branch(repo_path: str, branch: str):
    return subprocess.check_call(["git", "checkout", branch], cwd=repo_path)


def _force_push_changes_to_test_branch_of_pydash_repo(
    branch: str,
    repo_path: str,
):
    subprocess.check_call(
        ["git", "push", "--force-with-lease", "origin", branch],
        cwd=repo_path,
    )



"""enable_basic_auth"""

import os
import dash
from flask_basicauth import BasicAuth


def enable_basic_auth(dash_app: dash.Dash):
    """Turn on basic auth for the dash app if required"""
    if os.environ.get("STAGE") == "production":
        server = dash_app.server

        # Forcing the auth to prompt on any page if it hasn't already been done.
        server.config["BASIC_AUTH_FORCE"] = True

        server.config["BASIC_AUTH_USERNAME"] = os.environ.get("APP_USERNAME")
        server.config["BASIC_AUTH_PASSWORD"] = os.environ.get("APP_PASSWORD")
        BasicAuth(server)



"""Module containing Enums for Gov.UK colours and ONS Accessible Colours.

Contains:
- GovUKColours: From the GOV.UK colour scheme:
    https://design-system.service.gov.uk/styles/colour/
- ONSAccessibleColours: ONS colours taken from the Subnational Indicators Explorer.
"""
from enum import Enum


class GovUKColours(Enum):
    """
    From the GOV.UK colour scheme: https://design-system.service.gov.uk/styles/colour/
    """

    RED = "#d4351c"
    YELLOW = "#ffdd00"
    GREEN = "#00703c"
    BLUE = "#1d70b8"
    DARK_BLUE = "#003078"
    LIGHT_BLUE = "#5694ca"
    PURPLE = "#4c2c92"
    BLACK = "#0b0c0c"
    DARK_GREY = "#505a5f"
    MID_GREY = "#b1b4b6"
    LIGHT_GREY = "#f3f2f1"
    WHITE = "#ffffff"
    LIGHT_PURPLE = "#6f72af"
    BRIGHT_PURPLE = "#912b88"
    PINK = "#d53880"
    LIGHT_PINK = "#f499be"
    ORANGE = "#f47738"
    BROWN = "#b58840"
    LIGHT_GREEN = "#85994b"
    TURQUOISE = "#28a197"
    DLUHC_BLUE = "#092237"

    RED_LIGHT_TO_DARK = ["#fbebe8", "#ea9a8e", "#d4351c", "#6a1b0e", "#150503"]
    BLUE_LIGHT_TO_DARK = [
        "#e8f1f8",
        "#77a9d4",
        "#1d70b8",
        "#11436e",
        "#092237",
    ]


class ONSAccessibleColours(Enum):
    """ONS colours taken from their Subnational indicators explorer
    (https://www.ons.gov.uk/peoplepopulationandcommunity/wellbeing/articles/
    subnationalindicatorsexplorer/2022-01-06)

    Orange added to differentiate from CIPFA light blue
    """

    DARK_BLUE = "#206095"
    LIME = "#a8bd3a"
    MAROON = "#871a5b"
    # LIGHT_BLUE = "#27a0cc"
    ORANGE = "#f47738"



"""captioned_figure function"""
from typing import Any, Optional, Union
import plotly
import dash


def captioned_figure(
    figure: plotly.graph_objects.Figure,
    captions: Union[
        dash.development.base_component.Component,
        list[dash.development.base_component.Component],
    ],
    figure_name: str,
    *,
    graph_style: Optional[dict[str, Any]] = None,
    desktop_only: bool = False,
):
    """
    Return figure with attached caption that can be read by a screen reader.

    The caption will not be displayed to users viewing the website through a
    browser but is available for a screen reader to describe.

    The supplied graph and caption are wrapped in an HTML <figure> element.

    Args:
        figure (plotly.graph_objects.Figure): The plotly figure to display and caption.
        captions (Component, list[Component]): The captions to apply to the
            figure to be read by a screen reader.
        figure_name (str): Identifier for the figure. Should be lower case,
            with words separated by dashes, e.g. jitter-figure
        graph_style (Optional[dict], optional): Any custom style to apply to the graph.
            Defaults to None.
        desktop_only (bool, optional): Whether the figure should be replaced with
            the caption when viewed on mobile. Defaults to False.

    Returns:
        dash.html.Figure: Figure html element containing the graph and its caption.
    """
    if not graph_style:
        graph_style = {}

    if "height" not in graph_style.keys():
        graph_style["height"] = "450px"

    return dash.html.Figure(
        [
            dash.html.Div(
                children=dash.dcc.Graph(
                    id=f"{figure_name}-graph",
                    responsive=True,
                    figure=figure,
                    style=graph_style,
                    config={"displayModeBar": False},
                ),
                className="jitter-desktop-only" if desktop_only else "",
                **{"role": "img", "aria-labelledby": f"{figure_name}-caption"},
                id=f"{figure_name}-figure",
            ),
            dash.html.Figcaption(
                children=captions,
                id=f"{figure_name}-caption",
                className="govuk-visually-hidden-desktop-only"
                if desktop_only
                else "govuk-visually-hidden",
            ),
        ]
    )



"""apply_and_reset_filters_buttons"""
from dash import html


def apply_and_reset_filters_buttons():
    """
    Return clear all selections link which is aligned to the right
    """
    return html.Div(
        [
            html.A(
                "Reset filters",
                href="?",
                className="govuk-link govuk-body",
            ),
            html.Button(
                "Apply filters",
                id="submit-button",
                n_clicks=0,
                className="govuk-button govuk-!-margin-left-5",
            ),
        ],
        className="govuk-button-group",
        style={"float": "right"},
    )



"""notification_banner"""
from dash import html


def notification_banner(text: str):
    """
    Return Gov UK Design component notification banner component.
    """
    banner = html.Div(
        [
            html.Div(
                [
                    html.H2(
                        ["Important"],
                        className="govuk-notification-banner__title",
                        id="govuk-notification-banner-title",
                    )
                ],
                className="govuk-notification-banner__header",
            ),
            html.Div(
                [
                    html.P(
                        [
                            text,
                        ],
                        className="govuk-notification-banner__heading",
                    )
                ],
                className="govuk-notification-banner__content",
            ),
        ],
        className="govuk-notification-banner",
        role="region",
        **{
            "aria-labelledby": "govuk-notification-banner-title",
            "data-module": "govuk-notification-banner",
        }
    )
    return banner



"""navbar"""

from typing import Optional
from dash import html


def navbar(links, identifier: Optional[str] = None):
    """A navigation bar for switching between dashboards."""
    return html.Nav(
        html.Ul(links, className="moj-primary-navigation__list"),
        className="moj-primary-navigation",
        role="navigation",
        **{"aria-label": "Primary navigation"},
        id=identifier if identifier is not None else ""
    )


def navbar_link(text, href):
    """A link to another dashboard"""
    return html.Li(
        html.A(text, href=href, className="moj-primary-navigation__link"),
        className="moj-primary-navigation__item",
    )


def navbar_link_active(text, href):
    """
    A link to another dashboard that appears highlighted, suggesting to the user that they are
    already viewing the linked dashboard.
    """
    return html.Li(
        html.A(
            text,
            href=href,
            className="moj-primary-navigation__link",
            **{"aria-current": "page"}
        ),
        className="moj-primary-navigation__item",
    )



"""card_full_width"""
from dash import html


def card_full_width(children):
    """
    Apply CSS classes and styles to create a card with a grey background
    that fits the full width of its parent container using CSS flexbox.

    See https://developer.mozilla.org/en-US/docs/Learn/CSS/CSS_layout for more information.
    """
    return html.Div(children, className="mini-card", style={"flex": "1 1 100%"})



"""paragraph component"""
from enum import Enum
from typing import Union

from dash import html


class ParagraphSizes(str, Enum):
    """Enum class to specify the allowed paragraph sizes"""

    LEAD = "govuk-body-l"
    DEFAULT = "govuk-body"
    SMALL = "govuk-body-s"


def paragraph(
    children: Union[str, list], size: ParagraphSizes = ParagraphSizes.DEFAULT
) -> html.P:
    """
    Create a <p> Html component with the children provided

    Args:
        children (DataFrame): The children to render inside the dash paragraph component.
        size (str, list): The size of the text that this paragraph will display.
            Defaults to DEFAULT.

    Returns:
        html.P: The dash HTML object for a paragraph.

    Raises:
        ValueError: If the paragraph size supplied is not a valid size.
    """
    if size not in list(ParagraphSizes):
        raise ValueError(
            f"Size {size} is not a valid paragraph size (use ParagraphSize enum)."
        )

    return html.P(
        children,
        className=size,
    )



"""phase_banner"""
from dash import html


def phase_banner_with_feedback(
    phase: str, feedback_link: str, link_id: str = "feedback-link"
):
    """Creates a phase banner with a feedback link, which can be specified.
    The id of the 'feedback' link can be set to allow for targeting with callbacks.

    The phase banner is set out here as part of the Gov.UK design system:
    https://design-system.service.gov.uk/components/phase-banner/"""
    return html.Div(
        [
            html.P(
                [
                    html.Strong(
                        [phase],
                        className="govuk-tag govuk-phase-banner__content__tag",
                    ),
                    html.Span(
                        [
                            "This is a new service - your ",
                            html.A(
                                ["feedback"],
                                href=feedback_link,
                                className="govuk-link",
                                id=link_id,
                            ),
                            " will help us to improve it.",
                        ],
                        className="govuk-phase-banner__text",
                    ),
                ],
                className="govuk-phase-banner__content",
            )
        ],
        className="govuk-phase-banner",
    )



"""download_button"""
from dash import html


def download_button():
    """
    Return download button which is aligned to the right
    """
    return html.Div(
        [
            html.Button(
                "Download data",
                id="download-button",
                n_clicks=0,
                className="govuk-button govuk-button--secondary",
            ),
        ],
        className="govuk-button-group",
        style={"float": "right"},
    )



"""navbar"""

from typing import Optional
from dash import html, dcc


def side_navbar(links, identifier: Optional[str] = None):
    """A navigation bar for switching between dashboards."""
    return html.Nav(
        links,
        className="moj-side-navigation moj-side-navigation__list",
        role="navigation",
        id=identifier if identifier is not None else "",
    )


def side_navbar_link(text, href):
    """A link to another dashboard"""
    return html.Li(
        dcc.Link(text, href=href, className="govuk-link govuk-link--no-visited-state"),
        className="moj-side-navigation__item govuk-!-margin-bottom-1",
    )


def side_navbar_link_active(text, href):
    """
    A link to another dashboard that appears highlighted, suggesting to the user that they are
    already viewing the linked dashboard.
    """
    return html.Li(
        dcc.Link(text, href=href, className="govuk-link govuk-link--no-visited-state"),
        className=(
            "moj-side-navigation__item moj-side-navigation__item--active "
            "govuk-!-margin-bottom-1"
        ),
    )



"""Function for creating a table component from a dataframe"""
from typing import Optional
from pandas import DataFrame
from dash import html


def table_from_dataframe(
    dataframe: DataFrame,
    title: Optional[str] = None,
    first_column_is_header: bool = True,
    title_is_subtitle: bool = False,
    short_table: bool = True,
    **table_properties,
):
    """
    Displays a pandas DataFrame as a table formatted in the Gov.UK style

    Part of the Gov.UK Design System:
    https://design-system.service.gov.uk/components/table/


    Args:
        dataframe (DataFrame): Dataframe containing formatted data to display.
        title (str, optional): Title to display above the table. Defaults to None.
        first_column_is_header (bool, optional): Sets if the first column is a header column.
            Defaults to True.
        title_is_subtitle (bool, optional): Sets if the title should be displayed as a subtitle
            or full title. Defaults to False.
        short_table: (bool, optional): if False the header of the table will scroll with window.
        **table_properties: Any additional arguments for the html.Table object,
            such as setting a width or id.

    Returns:
        html.Table: The dash HTML object for the table.
    """
    table_contents = []

    if title:
        table_contents.append(
            html.Caption(
                title,
                className="govuk-table__caption govuk-table__caption--s"
                if title_is_subtitle
                else "govuk-table__caption govuk-table__caption--m",
            )
        )

    table_contents.append(
        html.Thead(
            html.Tr(
                [
                    html.Th(header, scope="col", className="govuk-table__header")
                    for header in dataframe.columns
                ],
                className="govuk-table__row",
            ),
            className="govuk-table__head-short" if short_table else "govuk-table__head",
        )
    )

    table_contents.append(
        html.Tbody(
            [
                html.Tr(
                    [html.Th(row[0], scope="row", className="govuk-table__header")]
                    + [html.Td(cell, className="govuk-table__cell") for cell in row[1:]]
                )
                if first_column_is_header
                else html.Tr(
                    [html.Td(cell, className="govuk-table__cell") for cell in row]
                )
                for _, row in dataframe.iterrows()
            ],
            className="govuk-table__body",
        )
    )

    return html.Table(table_contents, className="govuk-table", **table_properties)



""" Returns a dataframe after connecting to CDS, otherwise uses a csv already saved in the file"""
import os
import json
import pandas as pd
import pyodbc
import boto3


def get_data_from_cds_or_fallback_to_csv(
    cds_sql_query: str, csv_path: str, secret_name: str, cds_server_name: str
) -> pd.DataFrame:
    """Tries to return dataframe from CDS first via Pydash credentials,
    otherwise via Amazon WorkSpaces,
    otherwise via a file from folder.
    Inputs:
        cds_sql_query(str): SQL query string
        csv_path(str): Filepath for location of csv to fallback to
        secret_name(str): AWS Secrets Manager, secret name containing CDS credentials.
        cds_server_name(str): CDS Server name used in connection string
    Returns:
        pd.DataFrame
    """
    if (
        "DATA_FOLDER_LOCATION" in os.environ
        and os.environ["DATA_FOLDER_LOCATION"] == "tests/"
    ) or ("STAGE" in os.environ and os.environ["STAGE"] == "testing"):
        return pd.read_csv(csv_path)

    try:
        conn = pyodbc.connect(
            _get_pydash_connection_string(secret_name, cds_server_name)
        )
        print("Dataframe has been loaded from CDS using Pydash credentials")

        return pd.read_sql_query(
            cds_sql_query,
            conn,
        )

    except Exception as credential_error:  # pylint: disable=broad-except
        try:
            print(
                "Failed to load dataframe using Pydash credentials: ", credential_error
            )
            conn = pyodbc.connect(
                "Driver={SQL Server};"
                f"Server={cds_server_name};"
                "Database=Dashboards;"
                "Trusted_Connection=yes;"
            )
            print(
                "Dataframe has been loaded from CDS using Windows login authentication"
            )

            return pd.read_sql_query(
                cds_sql_query,
                conn,
            )

        except pyodbc.Error as conn_error_except:
            print(
                "Failed to load dataframe using Windows login authentication: ",
                conn_error_except,
            )
            print("Dataframe has been loaded from CSV")
            return pd.read_csv(csv_path)


def _get_pydash_connection_string(secret_name: str, cds_server_name: str):
    """
    Pydash aka DAP Hosting requires username and password
    Inputs:
        secret_name(str): AWS Secrets Manager, secret name containing CDS credentials.
        cds_server_name(str): CDS Server name used in connection string
    """
    credentials = _pydash_sql_credentials(secret_name)
    conn_string_dap = (
        "Driver={/usr/lib/libmsodbcsql-18.so};"
        f"Server={cds_server_name};"
        "TrustServerCertificate=yes;"
        "Database=Dashboards;"
    )
    return (
        f"{conn_string_dap}UID={credentials['username']};PWD={credentials['password']};"
    )


def _pydash_sql_credentials(secret_name: str):
    """
    Logging into CDS from Pydash requires user name and password.
    This method will return a dictionary containing the keys "username" and "password".
    Raises `botocore.exceptions.ClientError` if no credentials could be obtained
    Inputs:
        secret_name(str): AWS Secrets Manager, secret name containing CDS credentials.
    Returns:
        dict:   a dictionary containing the keys "username" and "password"
    """
    region_name = "eu-west-1"
    # Create a Secrets Manager client
    session = boto3.session.Session()
    client = session.client(service_name="secretsmanager", region_name=region_name)
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.get_secret_value

    get_secret_value_response = client.get_secret_value(SecretId=secret_name)

    secret = get_secret_value_response["SecretString"]

    credentials = json.loads(secret)
    return credentials
